// MSRItfUSBHID.cpp: Implementation of the CMSRItfUSBHID class.
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MSRItfUSBHID.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

extern int nReader;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMSRItfUSBHID::CMSRItfUSBHID(INTERFACE_TYPE emItfType){
//	ASSERT(emItfType==ITF_TYPE_USB_HID_KB || emItfType==ITF_TYPE_USB_HID);
	if(emItfType != ITF_TYPE_USB_HID_KB && emItfType != ITF_TYPE_USB_HID)		emItfType = ITF_TYPE_USB_HID; // Default interface is ITF_TYPE_USB_HID.

	m_emItfType = emItfType; // Store interface type as class member.
	m_bPortOpened = FALSE; // Close port during initialization.

	m_hMSRDev = INVALID_HANDLE_VALUE; // Initialize device handle.
	m_strMSRName.Empty(); // Initialize device name string.

	m_pPreprstData = NULL;
	m_pHidCaps = NULL;
	m_VendorIDTmp = NULL;
	m_USBHID_PIDTmp = NULL;
	m_MAGTUSBHID_PIDTmp = NULL;
	m_USBHIDKB_PIDTmp = NULL;
}

CMSRItfUSBHID::~CMSRItfUSBHID(){
	if(m_hMSRDev != NULL)		CloseMSRDevice(); // Close device.
}

//////////////////////////////////////////////////////////////////////
// Return value: TRUE, continue in loop to enumerate other devices. 
BOOL CMSRItfUSBHID::EnumMSRDevice2(INTERFACE_TYPE *pemItfType, ULONG ulIndex, CString& strDevName){
 	BOOL								bRet = FALSE;
	DWORD								dwErr = 0;

 	GUID								HidGuid;
 	SP_INTERFACE_DEVICE_DATA			DevInfoData;
	HDEVINFO							hHwDevInfo = INVALID_HANDLE_VALUE ;
    PSP_INTERFACE_DEVICE_DETAIL_DATA    pDetailData = NULL;
    ULONG                               ulPredictedLength = 0;
    ULONG                               ulRequiredLength = 0;
	HANDLE								hMSR = INVALID_HANDLE_VALUE;
	HIDD_ATTRIBUTES						HidAttr;
  
	//Initialization
	if(pemItfType != NULL)
		*pemItfType = ITF_TYPE_UNKNOWN;
	strDevName.Empty();

	//Get Guid of HID class
	HidD_GetHidGuid(&HidGuid);

	// Open a handle to the plug and play dev node.
	hHwDevInfo = SetupDiGetClassDevs (
                           &HidGuid,
                           NULL, // Define no enumerator (global)
                           NULL, // Define no
                           (DIGCF_PRESENT | // Only Devices present
                           DIGCF_INTERFACEDEVICE)); // Function class devices.

	if(hHwDevInfo == INVALID_HANDLE_VALUE)
	{
		 dwErr = GetLastError();
		 return FALSE;
	}

	DevInfoData.cbSize = sizeof (SP_INTERFACE_DEVICE_DATA);
	bRet = SetupDiEnumDeviceInterfaces (hHwDevInfo,
                                         0, // No care about specific PDOs
                                         &HidGuid,
                                         ulIndex,
                                         &DevInfoData);
	if(bRet == FALSE)
	{
		//bRet = FALSE;
		goto MY_EXIT;
	}

	//Continue to enumerate...
	bRet = TRUE;

	// allocate a function class device data structure to receive the
	// goods about this particular device.
	if( SetupDiGetInterfaceDeviceDetail (
				hHwDevInfo,
				&DevInfoData,
				NULL, // probing so no output buffer yet
				0, // probing so output buffer length of zero
				&ulRequiredLength,
				NULL ) == FALSE )				
	{
		dwErr = GetLastError();
		if(dwErr != ERROR_INSUFFICIENT_BUFFER)
			goto MY_EXIT;	

		ulPredictedLength = ulRequiredLength;
	} //end if


	pDetailData = (PSP_INTERFACE_DEVICE_DETAIL_DATA) malloc (ulPredictedLength);
//	ASSERT(pDetailData != NULL);
	pDetailData->cbSize = sizeof (SP_INTERFACE_DEVICE_DETAIL_DATA);

	// Retrieve the information from Plug and Play.
	if( SetupDiGetInterfaceDeviceDetail (
			   hHwDevInfo,
			   &DevInfoData,
			   pDetailData,
			   ulPredictedLength,
			   &ulRequiredLength,
			   NULL	) == FALSE )
	{
		goto MY_EXIT;
	}


	//open the MSR
	hMSR = CreateFile(	pDetailData->DevicePath,
						0, //GENERIC_READ | GENERIC_WRITE,
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						NULL, // no SECURITY_ATTRIBUTES structure
						OPEN_EXISTING, // No special create flags
						0, //FILE_FLAG_OVERLAPPED, // No special attributes
						NULL); // No template file
	if(hMSR == INVALID_HANDLE_VALUE)
		goto MY_EXIT;

	//Get Attributes
	memset(&HidAttr, 0, sizeof(HidAttr));
	if( HidD_GetAttributes(hMSR, &HidAttr) == FALSE )
		goto MY_EXIT;

	//if IDTECH and supported Product ID
	if((	HidAttr.VendorID == IDTECH_VENDOR_ID ) || (	HidAttr.VendorID == MagTek_VENDOR_ID ))
	{

		m_VendorIDTmp = HidAttr.VendorID;
		if((	HidAttr.VendorID == IDTECH_VENDOR_ID ) || (	HidAttr.VendorID == MagTek_VENDOR_ID ) ){
			if((HidAttr.ProductID == MOIR_USBHID_PID) || (HidAttr.ProductID == MOIR_USBHIDKB_PID)  || (HidAttr.ProductID == MOIR_USBHIDMAGTEK_PID) 
			|| (HidAttr.ProductID == SecureMOIR_USBHID_PID)  || (HidAttr.ProductID == SecureMOIR_USBHIDKB_PID) || (HidAttr.ProductID == SecureMOIR_USBHIDMAGTEK_PID) )
			
			{
			
				// MOIR USBHID (ID Tech) 
				if (HidAttr.ProductID == MOIR_USBHID_PID)
				{
					m_USBHID_PIDTmp = MOIR_USBHID_PID;
					nReader = 0;
				
				}
				
				// MOIR USBHID (MagTek)
				else if (HidAttr.ProductID == MOIR_USBHIDMAGTEK_PID)
				{
					m_USBHID_PIDTmp = MOIR_USBHIDMAGTEK_PID;
					nReader = 1;
				
				}

				// MOIR USBKB
				else if (HidAttr.ProductID == MOIR_USBHIDKB_PID)
				{
					m_USBHIDKB_PIDTmp = MOIR_USBHIDKB_PID;
					nReader = 2;
				
				}

				//  SecureMOIR USBHID (ID Tech)
				else if (HidAttr.ProductID == SecureMOIR_USBHID_PID)
				{
					m_USBHID_PIDTmp = SecureMOIR_USBHID_PID;
					nReader = 3;
				
				}
				
				// SecureMOIR USBHID (MagTek)
				else if (HidAttr.ProductID == SecureMOIR_USBHIDMAGTEK_PID)
				{
					m_USBHID_PIDTmp = SecureMOIR_USBHIDMAGTEK_PID;
					nReader = 4;
				
				}

				// SecureMOIR USBKB
				else if (HidAttr.ProductID == SecureMOIR_USBHIDKB_PID)
				{
					m_USBHIDKB_PIDTmp = SecureMOIR_USBHIDKB_PID;
					nReader = 5;
				
				}			
			}
			
		}

		if (HidAttr.ProductID == m_USBHIDKB_PIDTmp)
		{
			strDevName = pDetailData->DevicePath;
			if(pemItfType != NULL)
				*pemItfType = ITF_TYPE_USB_HID_KB;
		}
		else if( HidAttr.ProductID == m_USBHID_PIDTmp ) 
		{
			strDevName = pDetailData->DevicePath;
			if(pemItfType != NULL)
				*pemItfType = ITF_TYPE_USB_HID;
		}
	}

MY_EXIT:
	if(hMSR != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hMSR);
		hMSR = INVALID_HANDLE_VALUE;
	}
	if(hHwDevInfo != INVALID_HANDLE_VALUE)
	{
		SetupDiDestroyDeviceInfoList(hHwDevInfo);
		hHwDevInfo = INVALID_HANDLE_VALUE;
	}

	if(pDetailData != NULL)
	{
		free(pDetailData);
 		pDetailData = NULL;
	}	

	return bRet;
}


BOOL CMSRItfUSBHID::Index2Name(ULONG ulIndex, CString& strName)
{
 	BOOL	bRet = FALSE;
	DWORD	dwErr = 0;

 	GUID	HidGuid;
 	SP_INTERFACE_DEVICE_DATA			DevInfoData;
	HDEVINFO							hHwDevInfo = INVALID_HANDLE_VALUE ;
    PSP_INTERFACE_DEVICE_DETAIL_DATA    pDetailData = NULL;
    ULONG                               ulPredictedLength = 0;
    ULONG                               ulRequiredLength = 0;
  
	//Initialization
	strName.Empty();

	//Get Guid of HID class
	HidD_GetHidGuid(&HidGuid);

	// Open a handle to the plug and play dev node.
	hHwDevInfo = SetupDiGetClassDevs (
                           &HidGuid,
                           NULL, // Define no enumerator (global)
                           NULL, // Define no
                           (DIGCF_PRESENT | // Only Devices present
                           DIGCF_INTERFACEDEVICE)); // Function class devices.

	if(hHwDevInfo == INVALID_HANDLE_VALUE)
	{
		 dwErr = GetLastError();
		 return FALSE;
	}

	DevInfoData.cbSize = sizeof (SP_INTERFACE_DEVICE_DATA);
	bRet = SetupDiEnumDeviceInterfaces (hHwDevInfo,
                                         0, // No care about specific PDOs
                                         &HidGuid,
                                         ulIndex,
                                         &DevInfoData);
	if(bRet == FALSE)
	{
		//bRet = FALSE;
		goto MY_EXIT;
	}

	// allocate a function class device data structure to receive the
	// goods about this particular device.
	bRet = SetupDiGetInterfaceDeviceDetail (
				hHwDevInfo,
				&DevInfoData,
				NULL, // probing so no output buffer yet
				0, // probing so output buffer length of zero
				&ulRequiredLength,
				NULL); // not interested in the specific dev-node
	
	if(bRet == FALSE)
	{
		dwErr = GetLastError();
		if(dwErr != ERROR_INSUFFICIENT_BUFFER)
		{
			//bRet = FALSE;
			goto MY_EXIT;	
		}

		ulPredictedLength = ulRequiredLength;
	} //end if


	pDetailData = (PSP_INTERFACE_DEVICE_DETAIL_DATA) malloc (ulPredictedLength);
	pDetailData->cbSize = sizeof (SP_INTERFACE_DEVICE_DETAIL_DATA);

	// Retrieve the information from Plug and Play.
	bRet = SetupDiGetInterfaceDeviceDetail (
			   hHwDevInfo,
			   &DevInfoData,
			   pDetailData,
			   ulPredictedLength,
			   &ulRequiredLength,
			   NULL);

	if(bRet == TRUE)
		strName = pDetailData->DevicePath;

MY_EXIT:

	if(hHwDevInfo != INVALID_HANDLE_VALUE)
	{
		SetupDiDestroyDeviceInfoList(hHwDevInfo);
		hHwDevInfo = INVALID_HANDLE_VALUE;
	}

	if(pDetailData != NULL)
	{
		free(pDetailData);
 		pDetailData = NULL;
	}
	
	return bRet;
}


BOOL CMSRItfUSBHID::OpenMSRDevice( ULONG ulIndex )
{
	BOOL bRet = FALSE;
	CString strDevName;

	if(m_bPortOpened)
		return TRUE;

	m_ulIndex = ulIndex;

	//Index => DeviceName
	bRet = Index2Name(ulIndex, strDevName);
	if(bRet == FALSE || strDevName.IsEmpty())
		return FALSE;

	m_bPortOpened = OpenMSRDeviceByName(strDevName);
	return m_bPortOpened;
}

BOOL CMSRItfUSBHID::OpenMSRDeviceByName( LPCTSTR lpszName)
{
	BOOL bRet = FALSE;

	USHORT usPID=0, usPID2=0; //for supporting m_MAGTUSBHID_PIDTmp
	CString strDevName;
	DWORD dwAccess=0;
	HANDLE hMSR = INVALID_HANDLE_VALUE;
	HIDD_ATTRIBUTES HidAttr;
	NTSTATUS nStatus;
	
	if(lpszName==NULL || lstrlen(lpszName)==0)
		return FALSE;

	strDevName = lpszName;
	if(m_strMSRName.CompareNoCase(strDevName) == 0)
	{
//		ASSERT(m_hMSRDev != INVALID_HANDLE_VALUE);
		return TRUE;
	}
	else
	{
		if(m_hMSRDev != INVALID_HANDLE_VALUE)
			CloseMSRDevice();
	}

	//Interface Type and its desired access
	switch(m_emItfType) //Set outside!!!
	{
	case ITF_TYPE_USB_HID_KB:
		usPID = m_USBHIDKB_PIDTmp;
		dwAccess = 0;
		break;
	case ITF_TYPE_USB_HID:
		usPID = m_USBHID_PIDTmp;
		usPID2 = m_MAGTUSBHID_PIDTmp;
		dwAccess = GENERIC_READ | GENERIC_WRITE;
		break;
	}


	//open the MSR
	hMSR = CreateFile(	strDevName,
						dwAccess, //GENERIC_READ | GENERIC_WRITE,
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						NULL, // no SECURITY_ATTRIBUTES structure
						OPEN_EXISTING, // No special create flags
						0, //FILE_FLAG_OVERLAPPED, // No special attributes
						NULL); // No template file
	if(hMSR == INVALID_HANDLE_VALUE)
		return FALSE;

	//Get Attributes
	memset(&HidAttr, 0, sizeof(HidAttr));
	bRet = HidD_GetAttributes(hMSR, &HidAttr);
	if(bRet == FALSE)
	{	
		//bRet = FALSE;
		goto MY_EXIT;
	}

	//if IDTECH and supported Product ID
	if(((HidAttr.VendorID != IDTECH_VENDOR_ID) &&  (HidAttr.VendorID != MagTek_VENDOR_ID)) || (HidAttr.ProductID != usPID	&& HidAttr.ProductID != usPID2) )
	{
		bRet = FALSE;
		goto MY_EXIT;
	}

//	ASSERT(m_pPreprstData == NULL);
	bRet = HidD_GetPreparsedData(hMSR, &m_pPreprstData);
	if(bRet==FALSE || m_pPreprstData==NULL)
	{
		bRet = FALSE;
		goto MY_EXIT;
	}
	
//	ASSERT(m_pHidCaps == NULL);
	m_pHidCaps = (PHIDP_CAPS)calloc(1, sizeof(HIDP_CAPS));
	if(m_pHidCaps == NULL)
	{
		bRet = FALSE;
		goto MY_EXIT;
	}
	nStatus	= HidP_GetCaps(m_pPreprstData, m_pHidCaps);
	if(nStatus != HIDP_STATUS_SUCCESS)
	{
		bRet = FALSE;
		goto MY_EXIT;
	}
	
	//set related member data
	m_strMSRName = strDevName;
	m_hMSRDev = hMSR;

	//DON'T FORGET!!
	bRet = TRUE;
	
MY_EXIT:
	if(bRet == FALSE)
	{
		if(m_pHidCaps != NULL)
		{
			free(m_pHidCaps);
			m_pHidCaps = NULL;
		}

		if(m_pPreprstData != NULL)
		{
			HidD_FreePreparsedData(m_pPreprstData);
			m_pPreprstData = NULL;
		}

		if(hMSR != INVALID_HANDLE_VALUE)
		{
			CloseHandle(hMSR);
			hMSR = INVALID_HANDLE_VALUE;
		}
	}

	return bRet;
}

BOOL CMSRItfUSBHID::SendCommand(BYTE* pCmdBuf,ULONG cbCmdBuf,PBYTE pRetBuf,PULONG pcbRetBuf)  
{
#ifdef _DEBUG
//TRACE("\nSendCommand:");
for(ULONG n=0;n<cbCmdBuf;n++)
{
//	if(!(n%16))
//		TRACE("\n");
//	TRACE("%02x ",pCmdBuf[n]);
}
//TRACE("\n");
#endif
	BOOL bRet=FALSE, bNAKReport=FALSE, bFirstNonZeroBlock = FALSE;
	ULONG ulRptLen = 0;
	ULONG ulRptCnt = 0, i, j, k, counter = 0;
	PBYTE pRptBuf=NULL, p=NULL;
	ULONG ulRetLen=0;

	if(pcbRetBuf == NULL)
		return FALSE;

	if(m_pHidCaps == NULL || m_hMSRDev == INVALID_HANDLE_VALUE)
	{
		*pcbRetBuf = 0;
		return FALSE;
	}

	ulRptLen = m_pHidCaps->FeatureReportByteLength; 
	pRptBuf = (PBYTE)calloc(ulRptLen, sizeof(BYTE));
	if(pRptBuf == NULL)
	{
		*pcbRetBuf = 0;
		return FALSE;
	}

	//remove one byte occupied by ReportID
	ulRptLen--;
	ulRptCnt = (cbCmdBuf + ulRptLen - 1) / ulRptLen;
	p = pCmdBuf;

	//Send all the feature report
	for(i=0; i<ulRptCnt; i++)
	{
		p = pCmdBuf + i * ulRptLen;

		memset(pRptBuf, 0, ulRptLen+1);
		memcpy(	&pRptBuf[1], 
				p, 
				i==ulRptCnt-1 ? cbCmdBuf-i*ulRptLen : ulRptLen);

		//try 3 times
		for(j=0; j<3; j++)
		{
			bRet = HidD_SetFeature(m_hMSRDev, pRptBuf, ulRptLen+1);
			if(bRet)	break;
			Sleep(2); //2ms
		}
		if(j == 3)
		{
			bRet = FALSE;
			goto MY_EXIT;
		}
		
		Sleep(2);

	} //end for

	Sleep(500);

	//get response data from MSR
	for(k=0;k<100;k++) // 1 second in total
	{
		{
			bNAKReport = FALSE;
			memset(pRptBuf, 0, ulRptLen+1);
			bRet = HidD_GetFeature(m_hMSRDev, pRptBuf, ulRptLen+1);
	
			if(bRet)
			{
				//whether all the successive char are 0 or not,
				//if NO, it is still valid response data...
				for(i=0; i<ulRptLen; i++)
					if(pRptBuf[i] != 0x0)
					{
						bFirstNonZeroBlock = true;
						counter = 0;
						break;
					}	
			}

		}
		
		if(ulRetLen + ulRptLen > *pcbRetBuf)
		{			
			bRet = FALSE;
			goto MY_EXIT;
		}

		// Pack the data after first non zero data block 
		if(bFirstNonZeroBlock)
		{
			memcpy(&pRetBuf[ulRetLen], &pRptBuf[1], ulRptLen);
			ulRetLen += ulRptLen;
		}

		if(bFirstNonZeroBlock)		
		{
			Sleep(10);

			// Exit after the 3rd 0xE0 packages
			if (counter > 2)
				goto MY_EXIT;
		}

		else
		{
			Sleep(100);
		
			// Exit after the 21th 0xE0 packages
			if (counter > 20)
				goto MY_EXIT;

		}



	} //end for

MY_EXIT:

	*pcbRetBuf = ulRetLen;

	if(pRptBuf != NULL)
	{
		free(pRptBuf);
		pRptBuf = NULL;
	}

	return bRet;
}

BOOL CMSRItfUSBHID::CloseMSRDevice()
{
	if(m_pHidCaps != NULL)
	{
		free(m_pHidCaps);
		m_pHidCaps = NULL;
	}

	if(m_pPreprstData != NULL)
	{
		HidD_FreePreparsedData(m_pPreprstData);
		m_pPreprstData = NULL;
	}

	if(m_hMSRDev != INVALID_HANDLE_VALUE)
	{
		CloseHandle(m_hMSRDev);
		m_hMSRDev = INVALID_HANDLE_VALUE;
	}

	m_strMSRName.Empty();
	m_bPortOpened = FALSE;

	return TRUE;
}

BOOL CMSRItfUSBHID::Send(BYTE* pCmdBuf,ULONG cbCmdBuf)
{
#ifdef _DEBUG
//TRACE("\nSendCommand:");
for(ULONG n=0;n<cbCmdBuf;n++)
{
//	if(!(n%16))
//		TRACE("\n");
//	TRACE("%02x ",pCmdBuf[n]);
}
//TRACE("\n");
#endif
	BOOL bRet=FALSE, bNAKReport=FALSE;
	ULONG ulRptLen = 0;
	ULONG ulRptCnt = 0, i, j;
	PBYTE pRptBuf=NULL, p=NULL;
	ULONG ulRetLen=0;

	if(m_pHidCaps == NULL || m_hMSRDev == INVALID_HANDLE_VALUE)
	{
		return FALSE;
	}

	ulRptLen = m_pHidCaps->FeatureReportByteLength; //??
	pRptBuf = (PBYTE)calloc(ulRptLen, sizeof(BYTE));
	if(pRptBuf == NULL)
	{
		return FALSE;
	}

	//remove one byte occupied by ReportID
	ulRptLen--;
	ulRptCnt = (cbCmdBuf + ulRptLen - 1) / ulRptLen;
	p = pCmdBuf;

	//Send all the feature report
	for(i=0; i<ulRptCnt; i++)
	{
		p = pCmdBuf + i * ulRptLen;

		memset(pRptBuf, 0, ulRptLen+1);
		//pRptBuf[0] = BYTE(i); //ReportID: 0 or i??
		memcpy(	&pRptBuf[1], 
				p, 
				i==ulRptCnt-1 ? cbCmdBuf-i*ulRptLen : ulRptLen);

		//try 3 times
		for(j=0; j<3; j++)
		{
			bRet = HidD_SetFeature(m_hMSRDev, pRptBuf, ulRptLen+1);
			if(bRet)	break;
			Sleep(10); //10ms
		}
		if(j == 3)
		{
			bRet = FALSE;			
		}

	} //end for
		
	CloseMSRDevice();
	return bRet;
}


BOOL CMSRItfUSBHID::SendCommand_MOIR(BYTE* pCmdBuf,ULONG cbCmdBuf,PBYTE pRetBuf,PULONG pcbRetBuf)
{
#ifdef _DEBUG
//TRACE("\nSendCommand:");
for(ULONG n=0;n<cbCmdBuf;n++)
{
//	if(!(n%16))
//		TRACE("\n");
//	TRACE("%02x ",pCmdBuf[n]);
}
//TRACE("\n");
#endif
	BOOL bRet=FALSE, bNAKReport=FALSE, bFirstNonZeroBlock = FALSE;
	ULONG ulRptLen = 0;
	ULONG ulRptCnt = 0, i, j, k, counter = 0;
	PBYTE pRptBuf=NULL, p=NULL;
	ULONG ulRetLen=0;

	if(pcbRetBuf == NULL)
		return FALSE;

	if(m_pHidCaps == NULL || m_hMSRDev == INVALID_HANDLE_VALUE)
	{
		*pcbRetBuf = 0;
		return FALSE;
	}

	ulRptLen = m_pHidCaps->FeatureReportByteLength; 
	pRptBuf = (PBYTE)calloc(ulRptLen, sizeof(BYTE));
	if(pRptBuf == NULL)
	{
		*pcbRetBuf = 0;
		return FALSE;
	}

	//remove one byte occupied by ReportID
	ulRptLen--;
	ulRptCnt = (cbCmdBuf + ulRptLen - 1) / ulRptLen;
	p = pCmdBuf;

	//Send all the feature report
	for(i=0; i<ulRptCnt; i++)
	{
		p = pCmdBuf + i * ulRptLen;

		memset(pRptBuf, 0, ulRptLen+1);
		memcpy(	&pRptBuf[1], 
				p, 
				i==ulRptCnt-1 ? cbCmdBuf-i*ulRptLen : ulRptLen);

		//try 3 times
		for(j=0; j<3; j++)
		{
			bRet = HidD_SetFeature(m_hMSRDev, pRptBuf, ulRptLen+1);
			if(bRet)	break;
			Sleep(2); //2ms
		}
		if(j == 3)
		{
			bRet = FALSE;
			goto MY_EXIT;
		}

		Sleep(2);

	} //end for

		
	Sleep(500);

	//get response data from MSR
	for(k=0;k<100;k++) // 1 second in total
	{
		
		bNAKReport = FALSE;
		memset(pRptBuf, 0, ulRptLen+1);
		bRet = HidD_GetFeature(m_hMSRDev, pRptBuf, ulRptLen+1);

		if(bRet)
		{
			//whether all the successive char are 0 or not,
			//if NO, it is still valid response data...
			for(i=0; i<ulRptLen; i++)			
				if((pRptBuf[1] != 0xE0) || (pRptBuf[2] != 0)  || (pRptBuf[3] != 0) || (pRptBuf[4] != 0) 
				|| (pRptBuf[5] != 0)  || (pRptBuf[6] != 0) || (pRptBuf[7] != 0) || (pRptBuf[8] != 0))
				{
					bFirstNonZeroBlock = true;
					counter = 0;
					break;
				}	
				
		
			//  NAK package (E0 0 0 0 0 0 0 0) 									
			if ((pRptBuf[1] == 0xE0) && (pRptBuf[2] == 0)  && (pRptBuf[3] == 0) && (pRptBuf[4] == 0) 
				&& (pRptBuf[5] == 0)  && (pRptBuf[6] == 0) && (pRptBuf[7] == 0) && (pRptBuf[8] == 0))		
			{
				bNAKReport	= true;	
				counter++;				
			}
		
		}

		
		
		if(ulRetLen + ulRptLen > *pcbRetBuf)
		{			
			bRet = FALSE;
			goto MY_EXIT;
		}

		// Pack the data after first non zero data block 
		if(bFirstNonZeroBlock)
		{
			memcpy(&pRetBuf[ulRetLen], &pRptBuf[1], ulRptLen);
			ulRetLen += ulRptLen;
		}

		if(bFirstNonZeroBlock)		
		{
			Sleep(10);

			// Exit after the 3rd 0xE0 packages
			if (counter > 2)
				goto MY_EXIT;
		}

		else
		{
			Sleep(100);
		
			// Exit after the 21th 0xE0 packages
			if (counter > 20)
				goto MY_EXIT;

		}


	} //end for

MY_EXIT:

	*pcbRetBuf = ulRetLen-25;

	if(pRptBuf != NULL)
	{
		free(pRptBuf);
		pRptBuf = NULL;
	}

	return bRet;
}
